//------------------------------------------------------------------
//  netBlock.cpp - Copyright (C) Ton Plooy 2000.
//  Compile/link with:
//  'cl ipflt.cpp -link iphlpapi.lib advapi32.lib user32.lib'
//------------------------------------------------------------------
#include <stdlib.h>
#define  WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <fltdefs.h>

SERVICE_STATUS_HANDLE s_hSvc      = NULL;
SERVICE_STATUS        s_SvcStat   = { 0 };
LPSTR                 s_szSvcName = "netBlock";
LPSTR                 s_pRegRoot  = "SOFTWARE\\WDJ\\netBlock";
INTERFACE_HANDLE      s_hIf       = NULL;

//------------------------------------------------------------------
BOOL        Install(char *szSvcName);
BOOL        Uninstall(char *szSvcName);
BOOL        AddIpAddress(char *pArgs);
void        AddFilter(char cIO, char *pLoc, char *pRem, char *pPrt);
BYTE       *stoip(char *szStrIP, BYTE IP[4]);
VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv);
VOID WINAPI Handler(DWORD fdwControl);
LONG        GetRegStr(char *szRoot, char *szKey,
                      char *szDefault, char *szVal, int nValSize);
LONG        SetRegStr(char *szRoot, char *szKey, char *szVal);

//------------------------------------------------------------------

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst,
                      LPSTR pArgs, int nCmdShow)
{
    SERVICE_TABLE_ENTRY SvcEntry[2];

    if ((pArgs != NULL) && (strncmp(pArgs, "-i", 2) == 0)) {
        // Install service
        if (Install(s_szSvcName) == TRUE) 
            MessageBox(NULL, "Service installed","netBlock", MB_OK);
        else
            MessageBox(NULL, "Install failed", "netBlock", MB_OK);
    }
    else if ((pArgs != NULL) && (strncmp(pArgs, "-u", 2) == 0)) {
        // Uninstall service
        if (Uninstall(s_szSvcName) == TRUE)
            MessageBox(NULL, "Service removed", "netBlock", MB_OK);
        else
            MessageBox(NULL, "Uninstall failed", "netBlock", MB_OK);
        SetRegStr(s_pRegRoot, "Filter", "");
    }
    else if ((pArgs != NULL) && (strncmp(pArgs, "-a", 2) == 0)) {
        // Add a new ip-address/mask
        if (AddIpAddress(pArgs + 2) == TRUE)
            MessageBox(NULL, "IP address added", "netBlock", MB_OK);
    }
    else {
        // Service started
        SvcEntry[0].lpServiceName = s_szSvcName;
        SvcEntry[0].lpServiceProc = ServiceMain;
        SvcEntry[1].lpServiceName = NULL;
        SvcEntry[1].lpServiceProc = NULL;
        StartServiceCtrlDispatcher(SvcEntry);
    }
    return 0;
}

//------------------------------------------------------------------
//
//------------------------------------------------------------------

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
    int   rc, nPort, nNum, nErr, nStat, bDtr;
    BYTE  szMsg[256], localIP[4];
    DWORD Start, Stop;

    s_hSvc = RegisterServiceCtrlHandler(s_szSvcName, Handler);

    s_SvcStat.dwServiceType             = SERVICE_WIN32_OWN_PROCESS;
    s_SvcStat.dwCurrentState            = SERVICE_RUNNING;
    s_SvcStat.dwControlsAccepted        = SERVICE_ACCEPT_STOP;
    s_SvcStat.dwWin32ExitCode           = NO_ERROR;
    s_SvcStat.dwServiceSpecificExitCode = NO_ERROR;
    s_SvcStat.dwCheckPoint              = 0;
    s_SvcStat.dwWaitHint                = 0;
    SetServiceStatus(s_hSvc, &s_SvcStat);

    // Create interface and set up IP filter(s)
    PfCreateInterface( 0, PF_ACTION_FORWARD, PF_ACTION_FORWARD,
                       FALSE, TRUE, &s_hIf );

    char szAddr[512], szLocal[32], *s = szAddr, *e = szAddr, *r, *p;
    GetRegStr(s_pRegRoot, "LocalNIC", "", szLocal, sizeof(szLocal));
    GetRegStr(s_pRegRoot, "Filter"  , "", szAddr , sizeof(szAddr));

    // Walk all comma separated entries
    while (e != NULL) {
        if ((e = strchr(s, ',')) != NULL) *e = '\0';   // Mark end

        for (r = s + 1; *r != '\0' && *r == ' '; r++); // Skip space
        for (p = r + 1; *p != '\0' && *p != ' '; p++); // Go past r
        if (*p == ' ') *p++ = '\0';                    // End of r
        for (         ; *p != '\0' && *p == ' '; p++); // Skip space

        AddFilter(s[0], szLocal, r, p);

        if (s != NULL) s = e + 1; // Advance s to next entry
    }
    stoip(szLocal, localIP); //
    rc = PfBindInterfaceToIPAddress(s_hIf, PF_IPV4, localIP);

    return ;
}

//------------------------------------------------------------------

VOID WINAPI Handler(DWORD fdwControl)
{
    // State change notifications
    switch (fdwControl) {
    case SERVICE_CONTROL_STOP:
        // Remove filter interface and stop service
        PfUnBindInterface( s_hIf );
        PfDeleteInterface( s_hIf );
        s_SvcStat.dwCurrentState = SERVICE_STOPPED;
        SetServiceStatus(s_hSvc, &s_SvcStat);
        break;
    default:
        break;
    }
    return ;
}

//------------------------------------------------------------------

BOOL Install(char *szSvcName)
{
    SC_HANDLE hSCM, hService;
    char      szFilePath[256];

    // Open the Service Control Manager
    if ((hSCM = OpenSCManager(NULL, NULL,
                              SC_MANAGER_ALL_ACCESS)) == NULL) {
        return FALSE;
    }
    // Get the executable file path
    GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));
    // Create the service
    hService = CreateService(hSCM, szSvcName, szSvcName,
                             SERVICE_ALL_ACCESS,
                             SERVICE_WIN32_OWN_PROCESS,
                             SERVICE_AUTO_START,
                             SERVICE_ERROR_NORMAL,
                             szFilePath,
                             NULL, NULL, NULL, NULL, NULL);
    if (hService == NULL) {
        CloseServiceHandle(hSCM);
        return FALSE;
    }
    CloseServiceHandle(hService);
    CloseServiceHandle(hSCM);

    return TRUE;
}

//------------------------------------------------------------------

BOOL Uninstall(char *szSvcName)
{
    SC_HANDLE hSCM, hService;
    BOOL      bRc = FALSE;

    // Open the Service Control Manager
    if ((hSCM = OpenSCManager(NULL, NULL,
                              SC_MANAGER_ALL_ACCESS)) == NULL) {
        return FALSE;
    }
    if ((hService = OpenService(hSCM, szSvcName, DELETE)) != NULL) {
        if (DeleteService(hService)) {
            bRc = TRUE;
        }
        CloseServiceHandle(hService);
    }
    CloseServiceHandle(hSCM);

    return bRc;
}

//------------------------------------------------------------------

BOOL AddIpAddress(char *pArgs)
{
    LONG rc;
    // i|o <local ip address> <remote ip-address> [port]
    if (pArgs[0] != 'i' && pArgs[0] != 'o' && pArgs[0] != 'l') {
        MessageBox(NULL, "Set address arguments:\n"
            "-ai  <remote IP> [t|u<port>]  (set input filter)\n"
            "-ao <remote IP> [t|u<port>]  (set output filter)\n"
            "-al  <local IP>  (specify adapter)\n\n"
            "Local IP must be set once.", "netBlock", MB_OK);
        return FALSE;
    }
    if (pArgs[0] == 'l') {
        pArgs++;
        while (isspace(*pArgs)) pArgs++;
        rc = SetRegStr(s_pRegRoot, "LocalNIC", pArgs);
    }
    else {
        char szAddr[512];
        GetRegStr(s_pRegRoot, "Filter", "", szAddr, sizeof(szAddr));
        wsprintf(szAddr + strlen(szAddr), "%s%s",
                szAddr[0] ? "," : "", pArgs);
        rc = SetRegStr(s_pRegRoot, "Filter", szAddr);
    }
    return rc == ERROR_SUCCESS;
}

//------------------------------------------------------------------

void AddFilter(char cIO, char *pLocal, char *pRemote, char *pPort)
{
    PF_FILTER_DESCRIPTOR ipFlt;
    BYTE                 srcIP[4], destIP[4];

    memset( &ipFlt, 0, sizeof(ipFlt) );
    ipFlt.dwFilterFlags     = FD_FLAGS_NOSYN;
    ipFlt.dwRule            = 0;
    ipFlt.pfatType          = PF_IPV4;
    ipFlt.dwProtocol        = pPort[0] == 't' ? FILTER_PROTO_TCP :
                              pPort[0] == 'u' ? FILTER_PROTO_UDP :
                                                FILTER_PROTO_ANY ;
    ipFlt.fLateBound        = 0;
    ipFlt.wSrcPort          = FILTER_TCPUDP_PORT_ANY;
    ipFlt.wDstPort          = pPort[0] ? atoi(pPort + 1) :
                              FILTER_TCPUDP_PORT_ANY;
    ipFlt.wSrcPortHighRange = ipFlt.wSrcPort;
    ipFlt.wDstPortHighRange = ipFlt.wDstPort;
    if (cIO == 'i') {       // Input filter
        ipFlt.SrcAddr       = stoip(pRemote, destIP);
        ipFlt.SrcMask       = (PBYTE)"\xFF\xFF\xFF\xFF";
        ipFlt.DstAddr       = stoip(pLocal, srcIP);
        ipFlt.DstMask       = (PBYTE)"\xFF\xFF\xFF\xFF";
        PfAddFiltersToInterface( s_hIf, 1, &ipFlt, 0, NULL, NULL );
    }
    else {                  // Output filter
        ipFlt.SrcAddr       = stoip(pLocal, srcIP);
        ipFlt.SrcMask       = (PBYTE)"\xFF\xFF\xFF\xFF";
        ipFlt.DstAddr       = stoip(pRemote, destIP);
        ipFlt.DstMask       = (PBYTE)"\xFF\xFF\xFF\xFF";
        PfAddFiltersToInterface( s_hIf, 0, NULL, 1, &ipFlt, NULL );
    }

    return ;
}

//------------------------------------------------------------------

BYTE *stoip(char *szStrIP, BYTE IP[4])
{
    // String to byte IP
    char szStr[64], *p;
    strncpy(szStr, szStrIP, sizeof(szStr));
    szStr[sizeof(szStr) - 1] = '\0';
    if ((p = strrchr(szStr, '.')) != NULL) {
        *p = '\0', IP[3] = atoi(p + 1);
        if ((p = strrchr(szStr, '.')) != NULL) {
            *p = '\0', IP[2] = atoi(p + 1);
            if ((p = strrchr(szStr, '.')) != NULL) {
                *p = '\0';
                IP[1] = atoi(p + 1);
                IP[0] = atoi(szStr);
            }
        }
    }
    return IP;
}

//------------------------------------------------------------------

LONG GetRegStr(char *szRoot, char *szKey, char *szDefault,
               char *szVal, int nValSize)
{
    LONG rc; DWORD dwLen, dwType; HKEY  hKey;

    szVal[0]  = '\0';
    if ((rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szRoot, 0L,
                KEY_ALL_ACCESS, &hKey)) == ERROR_SUCCESS) {
        dwLen = (DWORD)nValSize;
        rc = RegQueryValueEx(hKey, szKey, 0L, &dwType,
                            (BYTE *)szVal, &dwLen);
        RegCloseKey(hKey);
    }
    if ((rc != ERROR_SUCCESS) && (szDefault != NULL))
        strncpy((char *)szVal, szDefault, nValSize);

    return rc;
}

//------------------------------------------------------------------

LONG SetRegStr(char *szRoot, char *szKey, char *szVal)
{
    LONG rc; DWORD dwDisp; HKEY hKey;

    if ((rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szRoot, 0L,
                    KEY_ALL_ACCESS, &hKey)) != ERROR_SUCCESS) {
        if ((rc = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szRoot, 0L,
                    NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                    NULL, &hKey, &dwDisp)) != ERROR_SUCCESS) {
            return rc;
        }
    }
    RegSetValueEx(hKey, szKey, 0, REG_SZ,
                  (PBYTE)szVal, strlen(szVal) + 1);
    RegCloseKey(hKey);

    return rc;
}


