#if !defined(_PUBLISHER_H)
#define _PUBLISHER_H

typedef int EventType;

template <class SUBSCRIBER>
class CEventPublisher
{
protected:
    class CSubscriberList // Nested class declaration
        :public CTypedPtrList <CPtrList, SUBSCRIBER *> {};
    CMap <int, int, CSubscriberList*, CSubscriberList*&> m_map;

    // Protected constructor prevents instantiation
    // as a standalone object.
    CEventPublisher() {}
    virtual ~CEventPublisher();
    void AddEvent (EventType e);

public:
    void Subscribe (SUBSCRIBER *s, EventType e);
    void UnSubscribe (SUBSCRIBER *s, EventType e);
    void UnSubscribeAll (SUBSCRIBER *s);
};

// The destructor empties the map and deletes
// any list objects found therein.
template <class SUBSCRIBER>
inline CEventPublisher <SUBSCRIBER>
::~CEventPublisher ()  
{
    POSITION pos = m_map.GetStartPosition ();
    while (pos) 
    {
        EventType event;
        CSubscriberList *list;
        m_map.GetNextAssoc (pos, event, list);
        delete list;
    }
    m_map.RemoveAll ();
}

// AddEvent creates a subscriber list 
// for the specified event
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::AddEvent (EventType e)  
{
    CSubscriberList *list = new CSubscriberList;
    m_map.SetAt (e, list);
}

// Subscribe adds a pointer to the subscriber
// to the list of subscribers of the specified event.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::Subscribe    (SUBSCRIBER *s, EventType e)  
{
    CSubscriberList *list;
    if (m_map.Lookup (e, list))
    {
        if (!list->Find (s))
            list->AddTail (s);
    }
}

// UnSubscribe removes the specified subscriber
// from the specified event subscribers list.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::UnSubscribe (SUBSCRIBER *s, EventType e) 
{
    CSubscriberList *list;
    if (m_map.Lookup (e, list))
    {
        POSITION pos = list->Find (s);
        if (pos != NULL)
            list->RemoveAt (pos);
    }
}

// UnSubscribe All removes the specified subscriber
// from all the subscriber lists.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::UnSubscribeAll (SUBSCRIBER *s)   
{
    POSITION pos = m_map.GetStartPosition ();
    while (pos) 
    {
        EventType event;
        CSubscriberList *list;
        m_map.GetNextAssoc (pos, event, list);
        UnSubscribe (s, event);
    }
}

// This macro implements the body of a publishing function
#define IMPLEMENT_PUBLISHER(EVENT, HANDLER) \
        CSubscriberList *list; \
        m_map.Lookup (EVENT, list); \
        POSITION pos = list->GetHeadPosition (); \
        while (pos) list->GetNext(pos)->HANDLER;

#endif

