#if !defined(_PUBLISH_H)
#define _PUBLISH_H

#pragma warning (disable:4100)
#pragma warning (disable:4786)
#include <map>

typedef int EventType;

template <class SUBSCRIBER>
class CEventPublisher
{
protected:
    typedef std::multimap <EventType, SUBSCRIBER> MapType;
    typedef MapType::iterator MapIterator;
    typedef std::pair <MapIterator, MapIterator> MapPair;

    MapType m_map;

    // FindSubscriber returns an iterator for
    // the (event, subscriber) pair if found
    MapIterator FindSubscriber (SUBSCRIBER s, EventType e)
    {
        MapPair range = m_map.equal_range(e);
        MapIterator mi = range.first;
        while (mi != range.second)
        {
            if ((*mi).second == s)
                return mi;
            mi++;
        }
        return m_map.end();
    }

public:
    void Subscribe (SUBSCRIBER s, EventType e);
    void UnSubscribe (SUBSCRIBER s, EventType e);
    void UnSubscribeAll (SUBSCRIBER s);
};

// Subscribe adds a pointer to the subscriber
// to the list of subscribers of the specified event.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::Subscribe (SUBSCRIBER s, EventType e)
{
    if (FindSubscriber (s, e) == m_map.end())
        m_map.insert (MapType::value_type (e,s));
}

// UnSubscribe removes the specified subscriber
// from the specified event subscribers list.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::UnSubscribe (SUBSCRIBER s, EventType e)
{
    MapIterator mi = FindSubscriber (s, e);
    if (mi != m_map.end())
        m_map.erase (mi);
}

// UnSubscribe All removes the specified subscriber
// from all the subscriber lists.
template <class SUBSCRIBER>
inline void CEventPublisher <SUBSCRIBER>
::UnSubscribeAll (SUBSCRIBER s)
{
    MapIterator mi = m_map.begin();
    while (mi != m_map.end())
    {
        MapIterator i = mi++;
        if ((*i).second == s)
            m_map.erase (i);
    }
}

// This macro implements the body of a publishing function
#define IMPLEMENT_PUBLISHER(EVENT, HANDLER) \
    MapPair range = m_map.equal_range(EVENT); \
    MapIterator mi = range.first; \
    while (mi != range.second) {(*mi++).second->HANDLER;}

#endif


