#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <tchar.h>
#include "wdjsub.h"

typedef struct SUBCLASSING {
    WNDPROC            wndProc;
    WNDPROC            wndProcSaved;
    void               *pData;
    struct SUBCLASSING *pNext;
} SUBCLASSING;

#define getWndProc( hwnd ) \
    ( (WNDPROC) GetWindowLong( hwnd, GWL_WNDPROC ) )
#define getHead( hwnd ) \
    ( (SUBCLASSING *) GetProp( hwnd, getPropertyName() ) )

static ATOM s_prop = 0;
#define PROP_NAME MAKEINTRESOURCE( s_prop )

LPCTSTR getPropertyName( void ) { return PROP_NAME; }

static BOOL allocPropertyName( void ) {
    s_prop = GlobalAddAtom( _T( "wdj_sub_root" ) );
    return 0 != s_prop;
}

static void releasePropertyName( void ) {
    GlobalDeleteAtom( s_prop );
}

static SUBCLASSING *find( WNDPROC wndProc, HWND hwnd ) {
    SUBCLASSING *p = getHead( hwnd );
    while ( 0 != p && wndProc != p->wndProc ) {
        p = p->pNext;
    }
    return p;
}

static BOOL setHead( HWND hwnd, const SUBCLASSING *pHead ) {
    if( 0 != GetProp( hwnd, getPropertyName() ) ) {
        RemoveProp( hwnd, getPropertyName() );
    }
    if ( 0 != pHead ) {
        SetProp( hwnd, getPropertyName(), (HANDLE) pHead );
    }
    return (HANDLE) pHead == GetProp( hwnd, getPropertyName() );
}

static BOOL initNode( HWND hwnd, SUBCLASSING *p, 
    WNDPROC wndProc, void *pData, SUBCLASSING *pNext ) 
{
    p->wndProc      = wndProc;
    p->pData        = pData;
    p->pNext        = pNext;
    p->wndProcSaved = SubclassWindow( hwnd, wndProc );
    return  0 != p->wndProcSaved;
}

BOOL wdjSubclass( WNDPROC wndProc, HWND hwnd, void *pData ) {
    SUBCLASSING *pHead = getHead( hwnd );
    SUBCLASSING *pThis = find( wndProc, hwnd );
    WNDPROC curr = getWndProc( hwnd );

    if ( 0 != pThis || curr == wndProc ) {
        return FALSE; //*** FUNCTION EXIT POINT
    }
    if ( allocPropertyName() ) {
        if ( 0 != (pThis = malloc( sizeof( *pThis ) ) ) ) {
            if ( initNode( hwnd, pThis, wndProc, pData, pHead ) ) {
                if ( setHead( hwnd, pThis ) ) {
                    return TRUE; //*** FUNCTION EXIT POINT
                }
                SubclassWindow( hwnd, pThis->wndProcSaved );
            }
            free( pThis );
        }
        releasePropertyName();
    }
    return FALSE;
}

BOOL wdjUnhook( WNDPROC id, HWND hwnd ) {
    SUBCLASSING *pHead = getHead( hwnd );
    SUBCLASSING *pThis = find( id, hwnd );
    const WNDPROC curr = getWndProc( hwnd );

    if ( 0 == pThis ) {
        return FALSE; //*** FUNCTION EXIT POINT
    }
    if ( pHead == pThis ) {
        if ( curr != id ) {
            return FALSE; //*** FUNCTION EXIT POINT
        }
        SubclassWindow( hwnd, pThis->wndProcSaved );
        setHead( hwnd, pThis->pNext );
    } else {
        SUBCLASSING *pPrev = pHead;
        while ( pPrev->pNext != pThis ) {
            pPrev = pPrev->pNext;
        }
        if ( pPrev->wndProcSaved != id ) {
            return FALSE; //*** FUNCTION EXIT POINT
        }
        pPrev->pNext        = pThis->pNext;
        pPrev->wndProcSaved = pThis->wndProcSaved;
    }
    free( pThis );
    releasePropertyName();
    return TRUE;
}

LRESULT wdjCallOldProc( 
    WNDPROC id, HWND hwnd, UINT msg, WPARAM w, LPARAM l )
{
    LRESULT lResult = 0;
    const SUBCLASSING *pThis = find( id, hwnd );
    if ( 0 != pThis ) {
        const WNDPROC oldProc = pThis->wndProcSaved;
        if ( WM_DESTROY == msg || WM_NCDESTROY == msg ) {
            wdjUnhook( id, hwnd );
        }
        if ( 0 != oldProc ) {
            lResult = CallWindowProc( oldProc, hwnd, msg, w, l );
        }
    }

    return lResult;
}

void *wdjGetData( WNDPROC id, HWND hwnd ) {
    const SUBCLASSING *pThis = find( id, hwnd );
    return 0 != pThis ? pThis->pData : 0;
}
